"""
The main entrypoint for the prefect flow
"""

from collections.abc import Generator
from datetime import datetime
from typing import Any, Set

from ftmq.model import Coverage
from prefect import flow, task
from prefect.task_runners import ConcurrentTaskRunner
from prefect_dask import DaskTaskRunner
from prefect_ray import RayTaskRunner

from investigraph import __version__, settings
from investigraph.model.context import BaseContext, Context
from investigraph.model.flow import Flow, FlowOptions
from investigraph.model.resolver import Resolver
from investigraph.util import data_checksum


def get_runner_from_env() -> ConcurrentTaskRunner | DaskTaskRunner | RayTaskRunner:
    if settings.TASK_RUNNER == "dask":
        return DaskTaskRunner
    if settings.TASK_RUNNER == "ray":
        return RayTaskRunner
    return ConcurrentTaskRunner


def get_task_cache_key(_, params) -> str:
    return params["ckey"]


@task(
    retries=settings.TASK_RETRIES,
    retry_delay_seconds=settings.TASK_RETRY_DELAY,
    cache_key_fn=get_task_cache_key,
    cache_expiration=settings.TASK_CACHE_EXPIRATION,
    refresh_cache=not settings.TASK_CACHE,
)
def aggregate(ctx: Context, results: list[str], ckey: str) -> Coverage:
    fragments, coverage = ctx.aggregate(ctx, results)
    ctx.log.info("AGGREGATED %d fragments to %d proxies", fragments, coverage.entities)
    ctx.log.info("OUTPUT: %s", ctx.config.load.entities_uri)
    return coverage


@task(
    retries=settings.TASK_RETRIES,
    retry_delay_seconds=settings.TASK_RETRY_DELAY,
    cache_key_fn=get_task_cache_key,
    cache_expiration=settings.TASK_CACHE_EXPIRATION,
    refresh_cache=not settings.TASK_CACHE,
)
def load(ctx: Context, ckey: str) -> str:
    proxies = ctx.cache.get(ckey)
    out = ctx.load_fragments(proxies, ckey=ckey)
    ctx.log.info("LOADED %d proxies", len(proxies))
    ctx.log.info("OUTPUT: %s", out)
    return out


@task(
    retries=settings.TASK_RETRIES,
    retry_delay_seconds=settings.TASK_RETRY_DELAY,
    cache_key_fn=get_task_cache_key,
    cache_expiration=settings.TASK_CACHE_EXPIRATION,
    refresh_cache=not settings.TASK_CACHE,
)
def transform(ctx: Context, ckey: str) -> str:
    proxies: list[dict[str, Any]] = []
    records = ctx.cache.get(ckey)
    for rec, ix in records:
        try:
            for proxy in ctx.config.transform.handle(ctx, rec, ix):
                proxies.append(proxy.to_dict())
        except Exception as e:
            ctx.log.error(f"{e.__class__.__name__}: {e}")
    ctx.log.info("TRANSFORMED %d records", len(records))
    return ctx.cache.set(proxies)


def extract(
    ctx: Context, ckey: str, res: Resolver | None = None
) -> Generator[str, None, None]:
    ctx.log.info("Starting EXTRACT stage ...")
    if res is not None:
        enumerator = enumerate(ctx.config.extract.handle(ctx, res), 1)
    else:
        enumerator = enumerate(ctx.config.extract.handle(ctx), 1)
    batch = []
    ix = 0
    for ix, rec in enumerator:
        batch.append((rec, ix))
        if ix % ctx.config.transform.chunk_size == 0:
            ctx.log.info("extracting record %d ...", ix)
            yield ctx.cache.set(batch)
            batch = []
    if batch:
        yield ctx.cache.set(batch)
    ctx.log.info("EXTRACTED %d records", ix)


@task(
    retries=settings.TASK_RETRIES,
    retry_delay_seconds=settings.TASK_RETRY_DELAY,
    cache_key_fn=get_task_cache_key,
    cache_expiration=settings.TASK_CACHE_EXPIRATION,
    refresh_cache=not settings.TASK_CACHE,
)
def extract_task(
    ctx: Context, ckey: str, res: Resolver | None = None
) -> Generator[str, None, None]:
    return extract(ctx, ckey, res)


def dispatch_extract(
    ctx: Context, ckey: str, res: Resolver | None = None
) -> Generator[str, None, None]:
    if ctx.config.extract.fetch:
        return extract_task(ctx, ckey, res)
    else:  # enable requests subflow
        return extract(ctx, ckey, res)


def run_pipeline(ctx: Context) -> list[Any]:
    res = None
    if ctx.config.extract.fetch:
        res = Resolver(source=ctx.source)
        if res.source.is_http:
            res._resolve_http()
        ckey = res.get_cache_key()
    else:
        ckey = ctx.source.uri

    results = []
    for key in dispatch_extract(ctx, f"extract-{ckey}", res):
        transformed = transform.submit(ctx, key)
        loaded = load.submit(ctx, transformed)
        results.append(loaded)

    return results


@flow(
    name="investigraph",
    version=__version__,
    flow_run_name="{options.flow_name}",
    task_runner=get_runner_from_env(),
)
def run(options: FlowOptions) -> Flow:
    flow = Flow.from_options(options)
    results = []
    ctxs: Set[Context] = set()
    ctx = BaseContext.from_config(flow.config)
    for source in ctx.config.seed.handle(ctx):
        ctxs.add(ctx.from_source(source))
    for source in ctx.config.extract.sources:
        ctxs.add(ctx.from_source(source))

    for ix, run_ctx in enumerate(ctxs):
        if ix == 0:  # only on first time
            ctx.export_metadata()
            ctx.log.info("INDEX: %s" % ctx.config.load.index_uri)
        results.extend(run_pipeline(run_ctx))

    if flow.config.aggregate:
        fragments = [r.result() for r in results]
        res = aggregate.submit(ctx, fragments, data_checksum(fragments))
        ctx.config.dataset.coverage = res.result()
        ctx.export_metadata()
        ctx.log.info("INDEX (updated with coverage): %s" % ctx.config.load.index_uri)

    flow.end = datetime.utcnow()
    fragment_uris = [r.result() for r in results]
    flow.fragment_uris = filter(lambda x: x is not None, fragment_uris)
    return flow



"""
Transform stage: map data records to ftm proxies
"""

from typing import TYPE_CHECKING

from investigraph.model.mapping import QueryMapping

if TYPE_CHECKING:
    from investigraph.model import Context

from ftmq.io import make_proxy

from investigraph.types import CEGenerator, SDict


def map_record(record: SDict, mapping: QueryMapping) -> CEGenerator:
    mapping = mapping.get_mapping()
    if mapping.source.check_filters(record):
        entities = mapping.map(record)
        for proxy in entities.values():
            yield make_proxy(proxy.to_dict())


def map_ftm(ctx: "Context", data: SDict, ix: int) -> CEGenerator:
    for mapping in ctx.config.transform.queries:
        yield from map_record(data, mapping)



from investigraph.model import Catalog, Context, Dataset, Resolver, Source, TaskContext
from investigraph.settings import VERSION as __version__

__all__ = [__version__, Catalog, Context, Dataset, TaskContext, Source, Resolver]



from functools import cache

from followthemoney import model
from followthemoney.mapping.query import QueryMapping as _QueryMapping
from ftmq.types import Properties, Schemata
from pydantic import BaseModel, Field


class PropertyMapping(BaseModel):
    column: str | None = None
    columns: list[str] | None = None
    join: str | None = None
    split: str | None = None
    entity: str | None = None
    format: str | None = None
    fuzzy: str | None = None
    required: bool | None = False
    literal: str | None = None
    literals: list[str] | None = None
    template: str | None = None


class EntityMapping(BaseModel):
    key: str | None = None
    keys: list[str] | None = []
    key_literal: str | None = None
    id_column: str | None = None
    schema_: Schemata = Field(..., alias="schema")
    properties: dict[Properties, PropertyMapping] = {}


class QueryMapping(BaseModel):
    entities: dict[str, EntityMapping] | None = {}

    def get_mapping(self) -> _QueryMapping:
        return load_mapping(self)

    def __hash__(self) -> int:
        return hash(repr(self.dict()))


@cache
def load_mapping(mapping: QueryMapping) -> _QueryMapping:
    mapping = mapping.dict(by_alias=True)
    mapping.pop("database", None)
    mapping["csv_url"] = "/dev/null"
    return model.make_mapping(mapping)



import shutil
import sys
from pathlib import Path
from typing import Annotated, Optional

import orjson
import typer
from ftmq.io import smart_write
from ftmq.model import Catalog
from prefect.settings import PREFECT_HOME
from rich import print

from investigraph.inspect import inspect_config, inspect_extract, inspect_transform
from investigraph.model.block import get_block
from investigraph.model.flow import FlowOptions
from investigraph.pipeline import run
from investigraph.settings import DATASETS_BLOCK, DATASETS_REPO, VERSION

cli = typer.Typer(no_args_is_help=True)


@cli.callback(invoke_without_command=True)
def cli_version(
    version: Annotated[Optional[bool], typer.Option(..., help="Show version")] = False
):
    if version:
        print(VERSION)
        raise typer.Exit()


@cli.command("run")
def cli_run(
    dataset: Annotated[Optional[str], typer.Option("-d")] = None,
    block: Annotated[Optional[str], typer.Option("-b")] = None,
    config: Annotated[Optional[str], typer.Option("-c")] = None,
    index_uri: Annotated[Optional[str], typer.Option(...)] = None,
    fragments_uri: Annotated[Optional[str], typer.Option(...)] = None,
    entities_uri: Annotated[Optional[str], typer.Option(...)] = None,
    aggregate: Annotated[Optional[bool], typer.Option(...)] = True,
    chunk_size: Annotated[Optional[int], typer.Option(...)] = None,
):
    """
    Execute a dataset pipeline
    """
    options = FlowOptions(
        dataset=dataset,
        block=block,
        config=config,
        index_uri=index_uri,
        fragments_uri=fragments_uri,
        entities_uri=entities_uri,
        aggregate=aggregate,
        chunk_size=chunk_size,
    )
    run(options)


@cli.command("add-block")
def cli_add_block(
    block: Annotated[
        str,
        typer.Option(
            "-b",
            prompt=f"Datasets configuration block, for example: {DATASETS_BLOCK}",
        ),
    ],
    uri: Annotated[
        str, typer.Option("-u", prompt=f"Block source uri, example: {DATASETS_REPO}")
    ],
):
    """
    Configure a datasets block (currently only github and local filesystem supported.)
    """
    block = get_block(block)
    try:
        block.register(uri)
        print(f"[bold green]OK[/bold green] block `{block}` created.")
    except ValueError as e:
        if "already in use" in str(e):
            print(f"[bold red]Error[/bold red] block `{block}` already existing.")
        else:
            raise e


@cli.command("inspect")
def cli_inspect(
    config_path: Annotated[Path, typer.Argument()],
    extract: Annotated[Optional[bool], typer.Option()] = False,
    transform: Annotated[Optional[bool], typer.Option()] = False,
    to_json: Annotated[Optional[bool], typer.Option()] = False,
):
    config = inspect_config(config_path)
    print(f"[bold green]OK[/bold green] `{config_path}`")
    print(f"[bold]dataset:[/bold] {config.dataset.name}")
    print(f"[bold]title:[/bold] {config.dataset.title}")

    if extract:
        for name, df in inspect_extract(config):
            print(f"[bold green]OK[/bold green] {name}")
            if to_json:
                for _, row in df.iterrows():
                    print(
                        orjson.dumps(
                            row.to_dict(), option=orjson.OPT_APPEND_NEWLINE
                        ).decode()
                    )
            else:
                print(df.to_markdown(index=False))

    if transform:
        for name, proxies in inspect_transform(config):
            print(f"[bold green]OK[/bold green] {name}")
            for proxy in proxies:
                data = orjson.dumps(
                    proxy.to_dict(), option=orjson.OPT_APPEND_NEWLINE
                ).decode()
                print(data)


@cli.command("build-catalog")
def cli_catalog(
    path: Annotated[Path, typer.Argument()],
    uri: Annotated[str, typer.Option("-o")] = "-",
    flatten: Annotated[Optional[bool], typer.Option(...)] = False,
):
    """
    Build a catalog from datasets metadata and write it to anywhere from stdout
    (default) to any uri `fsspec` can handle, e.g.:

        investigraph build-catalog catalog.yml -u s3://mybucket/catalog.json
    """
    catalog = Catalog.from_path(path)
    if uri != "-":
        catalog.uri = uri
    if flatten:
        datasets = [d.dict() for d in catalog.get_datasets()]
        data = {
            "datasets": datasets,
            "catalog": catalog.metadata(),
        }
    else:
        data = catalog.dict()
    data = orjson.dumps(data, option=orjson.OPT_APPEND_NEWLINE)
    if uri == "-":
        sys.stdout.write(data.decode())
    else:
        smart_write(uri, data)


@cli.command("reset")
def cli_reset(yes: Annotated[str, typer.Option(prompt="Are you sure? [yes/no]")]):
    """
    Reset all prefect data in `PREFECT_HOME`
    """
    if yes == "yes":
        path = PREFECT_HOME.value()
        shutil.rmtree(str(path), ignore_errors=True)
        print(f"[bold green]OK[/bold green] deleted everything in `{path}`.")



import mimetypes
from datetime import datetime
from urllib.parse import urlparse

import requests
from dateparser import parse as parse_date
from normality import slugify
from pantomime import normalize_mimetype, types
from pydantic import BaseModel
from runpandarun import Playbook
from runpandarun.util import PathLike, absolute_path

from investigraph.util import slugified_dict


class SourceHead(BaseModel):
    etag: str | None = None
    last_modified: datetime | None = None
    content_type: str | None = None
    content_length: int | None = None

    def __init__(self, **data):
        data = slugified_dict(data)
        last_modified = data.pop("last_modified", None)
        if isinstance(last_modified, str):
            data["last_modified"] = parse_date(last_modified)
        super().__init__(
            content_type=normalize_mimetype(data.pop("content_type", None)),
            **data,
        )

    def can_stream(self) -> bool:
        return self.content_type in (types.CSV, types.JSON)

    @property
    def ckey(self) -> str | None:
        if self.etag:
            return self.etag
        if self.last_modified:
            return self.last_modified.isoformat()


class Source(BaseModel):
    name: str
    uri: str
    scheme: str
    mimetype: str | None = None
    pandas: Playbook | None = Playbook()
    stream: bool | None = None

    def __init__(self, **data):
        data["uri"] = str(data["uri"])
        data["name"] = data.get("name", slugify(data["uri"]))
        data["scheme"] = data.get("scheme", urlparse(data["uri"]).scheme or "file")
        if "mimetype" not in data:
            mtype, _ = mimetypes.guess_type(data["uri"])
            data["mimetype"] = normalize_mimetype(mtype)
        data["stream"] = data.get("stream", data["mimetype"] == types.CSV)
        super().__init__(**data)

    def ensure_uri(self, base: PathLike) -> None:
        """
        ensure absolute file paths based on base path of paretn config.yml
        """
        if self.scheme.startswith("file"):
            self.uri = str(absolute_path(self.uri, base))

    @property
    def is_http(self) -> bool:
        return self.scheme.startswith("http")

    def head(self) -> SourceHead:
        if self.is_http:
            res = requests.head(self.uri)
            return SourceHead(**slugified_dict(res.headers))
        raise NotImplementedError("Cannot fetch head for scheme %s" % self.scheme)



from datetime import datetime
from typing import Any

from normality import slugify
from prefect.runtime import flow_run
from pydantic import BaseModel, root_validator

from investigraph.model.config import Config, get_config
from investigraph.settings import CHUNK_SIZE, DATA_ROOT, DATASETS_BLOCK
from investigraph.util import ensure_path


class FlowOptions(BaseModel):
    dataset: str | None = None
    block: str | None = None
    config: str | None = None
    aggregate: bool | None = None
    chunk_size: int | None = CHUNK_SIZE

    index_uri: str | None = None
    fragments_uri: str | None = None
    entities_uri: str | None = None

    @property
    def flow_name(self) -> str:
        if self.dataset is not None:
            return self.dataset
        return slugify(self.config)

    @root_validator
    def validate_options(cls, values):
        block = values.get("dataset") and (values.get("block") or DATASETS_BLOCK)
        config = values.get("config")
        if not block and not config:
            raise ValueError("Specify at least a config file or a block and dataset")
        return values


class Flow(BaseModel):
    dataset: str
    config: Config
    run_id: str | None = None
    start: datetime
    end: datetime | None = None
    fragment_uris: set[str] | None = set()
    entities_uri: str | None = None

    def __init__(self, **data):
        data["start"] = data.get("start", datetime.utcnow())
        data["run_id"] = data.get("run_id", flow_run.get_id())

        # override base config with runtime options
        options = data.pop("options", None)
        if options:
            config = get_config(
                data.pop("dataset", options.dataset),
                options.block,
                options.config,
            )

            self.assign(config.extract, "chunk_size", options.chunk_size)
            self.assign(config.transform, "chunk_size", options.chunk_size)
            self.assign(config.load, "chunk_size", options.chunk_size)

            self.assign(config.load, "index_uri", options.index_uri)
            self.assign(config.load, "fragments_uri", options.fragments_uri)
            self.assign(config.load, "entities_uri", options.entities_uri)
            if options.aggregate is False:
                config.aggregate = False
            if config.aggregate:
                self.assign(config.aggregate, "chunk_size", options.chunk_size)

            super().__init__(dataset=config.dataset.name, config=config, **data)
        else:
            super().__init__(**data)
        self.entities_uri = self.config.load.entities_uri
        path = ensure_path(DATA_ROOT / self.config.dataset.name)
        if self.config.load.index_uri is None:
            self.config.load.index_uri = (path / "index.json").as_uri()
        if self.config.load.fragments_uri is None:
            self.config.load.fragments_uri = (path / "fragments.json").as_uri()
        if self.config.load.entities_uri is None:
            self.config.load.entities_uri = (path / "entities.ftm.json").as_uri()

    @classmethod
    def from_options(cls, options: FlowOptions) -> "Flow":
        return cls(options=options)

    @staticmethod
    def assign(base: Any, attr: str, value: Any) -> None:
        if value is not None:
            setattr(base, attr, value)



"""
resolver for `.source.Source`
"""

from io import BytesIO

from ftmq.io import smart_open as open
from normality import slugify
from pantomime import types
from pydantic import BaseModel

from investigraph.exceptions import ImproperlyConfigured
from investigraph.logic import requests
from investigraph.model.source import Source, SourceHead
from investigraph.types import BytesGenerator
from investigraph.util import checksum

STREAM_TYPES = [types.CSV, types.JSON]


class Resolver(BaseModel):
    source: Source
    head: SourceHead | None = None
    response: requests.Response | None = None
    content: bytes | None = None
    checksum: str | None = None

    class Config:
        arbitrary_types_allowed = True

    @property
    def mimetype(self) -> str:
        if self.source.is_http:
            self._resolve_head()
            return self.head.content_type
        return self.source.mimetype

    @property
    def stream(self) -> bool:
        wants = self.source.stream
        if self.source.is_http:
            self._resolve_head()
            can = self.head.can_stream()
        else:
            can = self.mimetype in STREAM_TYPES
        if wants is None:
            return can
        return wants and can

    def _resolve_head(self) -> None:
        if self.head is None:
            self.head = self.source.head()

    def _resolve_http(self) -> None:
        if self.response is None:
            self._resolve_head()
            self.source.stream = self.source.stream or self.head.can_stream()
            res = requests.get(self.source.uri, stream=self.source.stream)
            assert res.ok
            self.response = res

    def _resolve_content(self) -> None:
        if self.content is None:
            if self.stream:
                raise ImproperlyConfigured("%s is a stream" % self.source.uri)
            if self.source.is_http:
                self._resolve_http()
                self.content = self.response.content
            else:
                with open(self.source.uri, "rb") as fh:
                    self.content = fh.read()
            self.checksum = checksum(BytesIO(self.content))

    def iter(self, chunk_size: int | None = 10_000) -> BytesGenerator:
        if self.source.stream:
            chunk = b""
            for ix, line in enumerate(self.iter_lines(), 1):
                chunk += line + b"\r"
                if ix % chunk_size == 0:
                    yield BytesIO(chunk)
                    chunk = b""
            if chunk:
                yield BytesIO(chunk)
        else:
            yield BytesIO(self.get_content())

    def iter_lines(self) -> BytesGenerator:
        if not self.source.stream:
            raise ImproperlyConfigured("%s is not a stream" % self.source.uri)

        if self.source.is_http:
            self._resolve_http()
            yield from self.response.iter_lines()
        else:
            with open(self.source.uri, "rb") as fh:
                yield from fh

    def get_content(self) -> bytes:
        self._resolve_content()
        return self.content

    def get_cache_key(self) -> str:
        slug = f"RESOLVE#{slugify(self.source.uri)}"
        if self.source.is_http:
            self._resolve_head()
            if self.head.ckey:
                return f"{slug}#{self.head.ckey}"
        if not self.source.stream:
            self._resolve_content()
            return self.checksum
        return slug  # handle expiration via cache_expiration on @task



from typing import TYPE_CHECKING, Callable

from banal import ensure_list, keys_values
from pydantic import BaseModel
from runpandarun import Playbook

from investigraph.model.mapping import QueryMapping
from investigraph.settings import (
    CHUNK_SIZE,
    DEFAULT_AGGREGATOR,
    DEFAULT_EXTRACTOR,
    DEFAULT_LOADER,
    DEFAULT_SEEDER,
    DEFAULT_TRANSFORMER,
    FTM_STORE_URI,
)
from investigraph.types import TaskResult
from investigraph.util import get_func, pydantic_merge

if TYPE_CHECKING:
    from investigraph.model.context import Context

from investigraph.model.source import Source


class Stage(BaseModel):
    _default_handler = None

    handler: str
    chunk_size: int | None = CHUNK_SIZE

    def __init__(self, **data):
        data["handler"] = data.pop("handler", self._default_handler)
        super().__init__(**data)

    def get_handler(self) -> Callable:
        return get_func(self.handler)

    def handle(self, ctx: "Context", *args, **kwargs) -> TaskResult:
        handler = self.get_handler()
        return handler(ctx, *args, **kwargs)


class SeedStage(Stage):
    _default_handler = DEFAULT_SEEDER

    glob: str | None = None


class ExtractStage(Stage):
    _default_handler = DEFAULT_EXTRACTOR

    fetch: bool | None = True
    sources: list[Source] | None = []
    pandas: Playbook | None = Playbook()

    def __init__(self, **data):
        super().__init__(**data)
        for source in self.sources:
            source.pandas = pydantic_merge(self.pandas, source.pandas)


class TransformStage(Stage):
    _default_handler = DEFAULT_TRANSFORMER

    queries: list[QueryMapping] | None = None

    def __init__(self, **data):
        data["queries"] = ensure_list(keys_values(data, "queries", "query"))
        super().__init__(**data)


class LoadStage(Stage):
    _default_handler = DEFAULT_LOADER

    index_uri: str | None = None
    fragments_uri: str | None = None
    entities_uri: str | None = None


class AggregateStage(Stage):
    _default_handler = DEFAULT_AGGREGATOR

    db_uri: str | None = FTM_STORE_URI

    def __init__(self, **data):
        if data.pop("handler", None) == "db":
            data["handler"] = "investigraph.logic.aggregate.in_db"
        super().__init__(**data)



import logging
from pathlib import Path

import yaml
from ftmq.model import Dataset
from ftmq.model.mixins import RemoteMixin, YamlMixin
from pydantic import BaseModel
from runpandarun.util import absolute_path

from investigraph.exceptions import ImproperlyConfigured
from investigraph.model.block import get_block
from investigraph.model.stage import (
    AggregateStage,
    ExtractStage,
    LoadStage,
    SeedStage,
    TransformStage,
)
from investigraph.settings import DATASETS_BLOCK
from investigraph.util import PathLike, is_module

log = logging.getLogger(__name__)


class Config(BaseModel, YamlMixin, RemoteMixin):
    dataset: Dataset
    base_path: Path | None = Path()
    seed: SeedStage | None = SeedStage()
    extract: ExtractStage | None = ExtractStage()
    transform: TransformStage | None = TransformStage()
    load: LoadStage | None = LoadStage()
    aggregate: bool | AggregateStage | None = AggregateStage()

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):
        if "dataset" not in data:
            data["dataset"] = data
        super().__init__(**data)
        # ensure absolute file paths for local sources
        self.base_path = Path(self.base_path).absolute()
        for source in self.extract.sources:
            source.ensure_uri(self.base_path)

    @classmethod
    def from_string(cls, data: str, base_path: PathLike | None = ".") -> "Config":
        data = yaml.safe_load(data)
        data["base_path"] = Path(base_path)
        config = cls(**data)

        # custom user code
        if not is_module(config.seed.handler):
            config.seed.handler = str(
                absolute_path(config.seed.handler, config.base_path)
            )
        if not is_module(config.extract.handler):
            config.extract.handler = str(
                absolute_path(config.extract.handler, config.base_path)
            )
        if not is_module(config.transform.handler):
            config.transform.handler = str(
                absolute_path(config.transform.handler, config.base_path)
            )
        if not is_module(config.load.handler):
            config.load.handler = str(
                absolute_path(config.load.handler, config.base_path)
            )
        if config.aggregate:
            if not is_module(config.aggregate.handler):
                config.aggregate.handler = str(
                    absolute_path(config.aggregate.handler, config.base_path)
                )

        return config


def get_config(
    dataset: str | None = None, block: str | None = None, path: PathLike | None = None
) -> Config:
    """
    Return configuration based on block or path (path has precedence)
    """
    if path is not None:
        return Config.from_path(path)
    if dataset is not None:
        block = block or DATASETS_BLOCK
        if block is not None:
            block = get_block(block)
            log.info("Using block `%s`" % block)
            block.load(dataset)
            return Config.from_path(block.path / dataset / "config.yml")
    raise ImproperlyConfigured("Specify `dataset` and `block` or `path` to config.")



"""
Load transformed data (aka CE proxies) to various targets like json files,
databases, s3 endpoints...
"""

from functools import cache
from typing import TYPE_CHECKING, Iterable, TypeAlias
from urllib.parse import urlparse

from ftmq.io import smart_write_proxies
from ftmstore import get_dataset

from investigraph.types import SDict

if TYPE_CHECKING:
    from investigraph.model import Context

TProxies: TypeAlias = Iterable[SDict]


def to_uri(uri: str, proxies: TProxies, **kwargs):
    smart_write_proxies(uri, proxies, **kwargs)


def to_store(uri: str, dataset: str, proxies: TProxies):
    dataset = get_dataset(dataset, database_uri=uri)
    bulk = dataset.bulk()
    for ix, proxy in enumerate(proxies):
        bulk.put(proxy, fragment=str(ix))
    bulk.flush()


class Loader:
    """
    write entity proxies to anywhere
    """

    def __init__(self, ctx: "Context", uri: str, parts: bool | None = False):
        self.uri = uri
        self.ctx = ctx
        self.is_store = "sql" in urlparse(uri).scheme
        self.parts = parts

    def write(self, proxies: TProxies, **kwargs) -> None:
        if self.is_store:
            to_store(self.uri, self.ctx.dataset, proxies)
            return self.uri
        else:
            uri = self.uri
            if self.parts:
                uri += f".{kwargs.pop('ckey')}"
            to_uri(uri, proxies, **kwargs)
            return uri


@cache
def get_loader(ctx: "Context", uri: str, parts: bool | None = False) -> Loader:
    return Loader(ctx, uri, parts)


def load_proxies(
    ctx: "Context", proxies: TProxies, uri: str, parts: bool | None = False, **kwargs
) -> str:
    loader = get_loader(ctx, uri, parts)
    return loader.write(proxies, **kwargs)



from ftmq.model import Catalog, Dataset

from investigraph.model.block import DatasetBlock
from investigraph.model.config import Config
from investigraph.model.context import Context, TaskContext
from investigraph.model.flow import Flow, FlowOptions
from investigraph.model.resolver import Resolver
from investigraph.model.source import Source

__all__ = [
    Dataset,
    DatasetBlock,
    Catalog,
    Config,
    Context,
    Flow,
    FlowOptions,
    Resolver,
    Source,
    TaskContext,
]



from typing import Any, Generator, TypeAlias

from ftmq.types import PathLike
from nomenklatura.entity import CE

# a string-keyed dict
Record: TypeAlias = dict[str, Any]
SDict: TypeAlias = Record  # FIXME backwards compatibility
RecordGenerator: TypeAlias = Generator[Record, None, None]

# composite entity generator
CEGenerator: TypeAlias = Generator[CE, None, None]

StrGenerator: TypeAlias = Generator[str, None, None]
BytesGenerator: TypeAlias = Generator[bytes, None, None]

TaskResult: TypeAlias = Generator[Any, None, None]
PathLike: TypeAlias = PathLike



"""
Dataset configuration (metadata and optional python function) is stored in
prefect blocks for easy distributed access
"""

from functools import cache, cached_property
from pathlib import Path

from normality import slugify
from prefect import filesystems
from prefect.blocks.core import Block

from investigraph.exceptions import BlockError
from investigraph.settings import DATA_ROOT
from investigraph.util import ensure_path

BLOCK_TYPES = ("github", "local-file-system")


class DatasetBlock:
    def __init__(self, prefix: str, name: str):
        if prefix not in BLOCK_TYPES:
            raise ValueError(f"unsupported block type: `{prefix}`")
        self.prefix = prefix
        self.name = name

    def __str__(self) -> str:
        return f"{self.prefix}/{self.name}"

    @cached_property
    def block(self) -> Block:
        return Block.load(str(self))

    @property
    def is_registered(self) -> bool:
        try:
            self.block
            return self.get_path().exists()
        except ValueError:
            return False

    @property
    def path(self) -> Path:
        if not self.is_registered:
            raise BlockError(f"Block `{self}` not registered yet.")
        return self.get_path()

    def get_path(self) -> Path:
        return DATA_ROOT / "blocks" / slugify(self.block._block_document_id, sep="_")

    def load_dataset(self, dataset: str) -> None:
        raise NotImplementedError  # subclass

    @classmethod
    def from_string(cls, block: str) -> "DatasetBlock":
        prefix, name = block.split("/")
        return cls(prefix=prefix, name=name)

    @staticmethod
    def get_create_kwargs(uri: str) -> dict[str, str]:
        raise NotImplementedError

    def register(
        self,
        uri: str,
        overwrite: bool | None = False,
        ignore_errors: bool | None = True,
    ) -> None:
        kwargs = self.get_create_kwargs(uri)
        block = self._block_cls(**kwargs)
        try:
            block.save(self.name, overwrite=overwrite)
        except ValueError as e:
            if not ignore_errors:
                raise e

    def load(self, dataset: str) -> None:
        """
        Load dataset from block and add path to python path
        """
        self.load_dataset(dataset)


class LocalFileSystemBlock(DatasetBlock):
    _block_cls = filesystems.LocalFileSystem

    @staticmethod
    def get_create_kwargs(uri: str) -> dict[str, str]:
        path = Path(uri).absolute()
        return {"basepath": path}

    def load_dataset(self, dataset: str) -> None:
        remote_path = Path(self.block.basepath) / dataset
        local_path = ensure_path(self.get_path() / dataset)
        self.block.get_directory(remote_path, local_path)


class GitHubBlock(DatasetBlock):
    _block_cls = filesystems.GitHub

    @staticmethod
    def get_create_kwargs(uri: str) -> dict[str, str]:
        return {"repository": uri, "include_git_objects": False}

    def load_dataset(self, dataset: str) -> None:
        self.block.get_directory(dataset, self.get_path())


@cache
def get_block(name: str, uri: str | None = None, **kwargs) -> DatasetBlock:
    if name.startswith("local"):
        block = LocalFileSystemBlock.from_string(name)
    elif name.startswith("github"):
        block = GitHubBlock.from_string(name)
    else:
        raise NotImplementedError(name)

    if uri is not None:
        block.register(uri, **kwargs)
    return block



import logging
from collections.abc import Iterable
from functools import cache
from typing import Any, Set

import fakeredis
import redis
from cachelib.serializers import RedisSerializer

from investigraph import settings
from investigraph.util import data_checksum

log = logging.getLogger(__name__)


class Cache:
    """
    This is an extremely simple cache interface for sharing tasks data
    efficiently via redis (or fakeredis during development)

    it creates (prefixed) keys based on input data

    it mimics redis GETDEL so that after fetching data from cache the key is
    deleted (turn of by `delete=False`)
    """

    serializer = RedisSerializer()

    def __init__(self):
        if settings.DEBUG:
            con = fakeredis.FakeStrictRedis()
            con.ping()
            log.info("Redis connected: `fakeredis`")
        else:
            con = redis.from_url(settings.REDIS_URL)
            con.ping()
            log.info("Redis connected: `{settings.REDIS_URL}`")
        self.cache = con

    def set(self, data: Any, key: str | None = None) -> str:
        key = key or data_checksum(data)
        data = self.serializer.dumps(data)
        self.cache.set(self.get_key(key), data)
        return key

    def get(self, key: str, delete: bool | None = settings.DEBUG) -> Any:
        key = self.get_key(key)
        res = self.cache.get(key)
        if delete:
            self.cache.delete(key)  # GETDEL
        if res is not None:
            data = self.serializer.loads(res)
            return data

    def sadd(self, *values: Iterable[Any], key: str | None = None) -> str:
        values = [str(v) for v in values]
        key = key or data_checksum(values)
        self.cache.sadd(self.get_key(key) + "#SET", *values)
        return key

    def smembers(self, key: str, delete: bool | None = settings.DEBUG) -> Set[str]:
        key = self.get_key(key) + "#SET"
        res: Set[bytes] = self.cache.smembers(key)
        if delete:
            self.cache.delete(key)
        return {v.decode() for v in res} or None

    def flushall(self):
        return self.cache.flushall()

    @staticmethod
    def get_key(key: str) -> str:
        return f"{settings.REDIS_PREFIX}:{key}"


@cache
def get_cache() -> Cache:
    return Cache()



from datetime import datetime
from typing import Iterable

import orjson
from followthemoney.util import make_entity_id
from ftmq.io import smart_write
from nomenklatura.entity import CE
from prefect import get_run_logger
from prefect.logging.loggers import PrefectLogAdapter
from pydantic import BaseModel

from investigraph.cache import Cache, get_cache
from investigraph.logic.aggregate import AggregatorResult, merge
from investigraph.model.config import Config
from investigraph.model.source import Source
from investigraph.types import CEGenerator
from investigraph.util import join_slug, make_proxy


class BaseContext(BaseModel):
    dataset: str
    prefix: str
    config: Config

    def __hash__(self) -> int:
        return hash(repr(self.dict()))

    def __eq__(self, other) -> bool:
        return hash(self) == hash(other)

    @property
    def cache(self) -> Cache:
        return get_cache()

    @property
    def log(self) -> PrefectLogAdapter:
        return get_run_logger()

    def load_fragments(self, *args, **kwargs) -> str:
        kwargs["uri"] = kwargs.pop("uri", self.config.load.fragments_uri)
        kwargs["parts"] = True
        return self.config.load.handle(self, *args, **kwargs)

    def load_entities(self, *args, **kwargs) -> str:
        kwargs["uri"] = kwargs.pop("uri", self.config.load.entities_uri)
        kwargs["parts"] = False
        return self.config.load.handle(self, *args, **kwargs)

    def aggregate(self, *args, **kwargs) -> AggregatorResult:
        return self.config.aggregate.handle(*args, **kwargs)

    def export_metadata(self) -> None:
        data = self.config.dataset
        data.updated_at = data.updated_at or datetime.utcnow()
        data = orjson.dumps(data.dict())
        smart_write(self.config.load.index_uri, data)

    def make_proxy(self, *args, **kwargs) -> CE:
        return make_proxy(*args, dataset=self.dataset, **kwargs)

    def make(self, *args, **kwargs) -> CE:
        # align with zavod api for easy migration
        return self.make_proxy(*args, **kwargs)

    def make_slug(self, *args, **kwargs) -> str:
        prefix = kwargs.pop("prefix", self.prefix)
        return join_slug(*args, prefix=prefix, **kwargs)

    def make_id(self, *args, **kwargs) -> str:
        prefix = kwargs.pop("prefix", self.prefix)
        return join_slug(make_entity_id(*args), prefix=prefix)

    def make_cache_key(self, *args: Iterable[str]) -> str:
        return join_slug(*args, sep="#")

    def task(self) -> "TaskContext":
        return TaskContext(**self.dict())

    def emit(self) -> None:
        raise NotImplementedError

    def from_source(self, source: Source) -> "Context":
        return Context(
            dataset=self.config.dataset.name,
            prefix=self.config.dataset.prefix,
            config=self.config,
            source=source,
        )

    @classmethod
    def from_config(cls, config: Config) -> "BaseContext":
        return cls(
            dataset=config.dataset.name,
            prefix=config.dataset.prefix,
            config=config,
        )


class Context(BaseContext):
    source: Source


class TaskContext(Context):
    proxies: dict[str, CE] = {}

    def __iter__(self) -> CEGenerator:
        yield from self.proxies.values()

    def emit(self, proxy: CE) -> None:
        # mimic zavod api, do merge already
        if proxy.id in self.proxies:
            self.proxies[proxy.id] = merge(self, self.proxies[proxy.id], proxy)
        else:
            self.proxies[proxy.id] = proxy


def init_context(config: Config, source: Source) -> Context:
    return Context(
        dataset=config.dataset.name,
        prefix=config.dataset.prefix,
        config=config,
        source=source,
    )



"""
Extract sources to iterate objects to dict records
"""

import numpy as np
import pandas as pd
from pantomime import types
from runpandarun.io import guess_handler_from_mimetype

from investigraph.model.context import Context
from investigraph.model.resolver import Resolver
from investigraph.types import RecordGenerator

TABULAR = [types.CSV, types.EXCEL, types.XLS, types.XLSX]
JSON = types.JSON


def yield_pandas(df: pd.DataFrame) -> RecordGenerator:
    for _, row in df.iterrows():
        yield dict(row.replace(np.nan, None))


def extract_pandas(
    resolver: Resolver, chunk_size: int | None = 10_000
) -> RecordGenerator:
    play = resolver.source.pandas
    play.read.uri = resolver.source.uri
    if play.read.handler is None:
        play.read.handler = f"read_{guess_handler_from_mimetype(resolver.mimetype)}"
    for ix, chunk in enumerate(resolver.iter(chunk_size)):
        df = play.read.handle(chunk)
        if resolver.mimetype == types.CSV:
            if ix == 0:  # first chunk
                play.read.options["names"] = df.columns
                play.read.options["skiprows"] = 0
                play.read.options["header"] = 0
        df = play.run(df)
        yield from yield_pandas(df)


# entrypoint
def handle(ctx: Context, resolver: Resolver, *args, **kwargs) -> RecordGenerator:
    yield from extract_pandas(resolver, *args, **kwargs)



"""
aggregate fragments
"""

from typing import TYPE_CHECKING, Literal, TypeAlias
from uuid import uuid4

from followthemoney import model
from ftmq.io import smart_read_proxies
from ftmq.model.coverage import Collector
from ftmstore import get_dataset

if TYPE_CHECKING:
    from investigraph.model import Context

from ftmq.model import Coverage

from investigraph.types import CE, CEGenerator

COMMON_SCHEMAS = ("Organization", "LegalEntity")

AggregatorResult: TypeAlias = tuple[int, Coverage]


def merge(ctx: "Context", p1: CE, p2: CE) -> CE:
    try:
        p1 = p1.merge(p2)
        p1.schema = model.common_schema(p1.schema, p2.schema)
        return p1
    except Exception as e:
        # try common schemata, this will probably "downgrade" entities
        # as in, losing some schema specific properties
        for schema in COMMON_SCHEMAS:
            if p1.schema.is_a(schema) and p2.schema.is_a(schema):
                p1 = ctx.make(schema, **p1.to_dict()["properties"])
                p1.id = p2.id
                p2 = ctx.make(schema, **p2.to_dict()["properties"])
                p1 = p1.merge(p2)
                return p1

        ctx.log.warn(f"{e}, id: `{p1.id}`")
        return p1


class Aggregator:
    """
    Aggregate loaded fragments to entities
    """

    def __init__(self, ctx: "Context", fragment_uris: list[str]) -> None:
        self.ctx = ctx
        self.fragment_uris = fragment_uris
        self.fragments = 0

    def get_fragments(self) -> CEGenerator:
        ix = -1
        for ix, proxy in enumerate(smart_read_proxies(self.fragment_uris)):
            if ix % self.ctx.config.aggregate.chunk_size == 0:
                self.ctx.log.info("reading in proxy %d ..." % ix)
            yield proxy
        self.fragments = ix + 1

    def aggregate_db(self) -> CEGenerator:
        dataset = get_dataset(
            "aggregate_%s" % uuid4().hex, database_uri=self.ctx.config.aggregate.db_uri
        )
        bulk = dataset.bulk()
        for ix, proxy in enumerate(self.get_fragments()):
            bulk.put(proxy, fragment=str(ix))
        bulk.flush()
        yield from dataset.iterate()
        dataset.drop()

    def aggregate_memory(self) -> CEGenerator:
        buffer = {}
        for proxy in self.get_fragments():
            if proxy.id in buffer:
                buffer[proxy.id] = merge(self.ctx, buffer[proxy.id], proxy)
            else:
                buffer[proxy.id] = proxy
        yield from buffer.values()

    def iterate(
        self, collector: Collector, handler: Literal["memory", "db"] | None = "memory"
    ) -> CEGenerator:
        aggregator = self.aggregate_db if handler == "db" else self.aggregate_memory
        for proxy in aggregator():
            collector.collect(proxy)
            yield proxy


def in_memory(ctx: "Context", fragment_uris: list[str]) -> AggregatorResult:
    aggregator = Aggregator(ctx, fragment_uris)
    with ctx.config.dataset.coverage as collector:
        proxies = aggregator.iterate(collector, "memory")
    ctx.load_entities(proxies, serialize=True)
    return aggregator.fragments, collector.export()


def in_db(ctx: "Context", fragment_uris: list[str]) -> AggregatorResult:
    aggregator = Aggregator(ctx, fragment_uris)
    with ctx.config.dataset.coverage as collector:
        proxies = aggregator.iterate(collector, "db")
    ctx.load_entities(proxies, serialize=True)
    return aggregator.fragments, collector.export()



import hashlib
import os
import re
from functools import cache, lru_cache
from importlib import import_module
from importlib.util import module_from_spec, spec_from_file_location
from io import BytesIO
from pathlib import Path
from typing import Any, Callable

import orjson
from banal import clean_dict, ensure_dict, is_listish
from followthemoney.util import join_text as _join_text
from followthemoney.util import make_entity_id, sanitize_text
from ftmq.util import make_dataset
from nomenklatura.dataset import DefaultDataset
from nomenklatura.entity import CE, CompositeEntity
from normality import slugify
from pydantic import BaseModel
from runpandarun.util import PathLike

from investigraph.exceptions import ImproperlyConfigured
from investigraph.types import SDict


def slugified_dict(data: dict[Any, Any]) -> SDict:
    return {slugify(k, "_"): v for k, v in ensure_dict(data).items()}


def make_proxy(schema: str, dataset: str | None = DefaultDataset, **properties) -> CE:
    data = {"schema": schema, "properties": properties}
    if isinstance(dataset, str):
        dataset = make_dataset(dataset)
    return CompositeEntity.from_data(dataset, data)


@cache
def ensure_path(path: PathLike) -> Path:
    path = Path(os.path.normpath(path))
    path.mkdir(parents=True, exist_ok=True)
    return path.absolute()


module_re = re.compile(r"^[\w\.]+:[\w]+")


@cache
def is_module(path: str) -> bool:
    return bool(module_re.match(path))


@cache
def get_func(path: str) -> Callable:
    module, func = path.rsplit(":", 1)
    if is_module(path):
        module = import_module(module)
    else:
        path = Path(module)
        spec = spec_from_file_location(path.stem, module)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
    return getattr(module, func)


@lru_cache(1024)
def clean_string(value: Any) -> str | None:
    """
    Convert a value to None or a sanitized string without linebreaks
    """
    value = sanitize_text(value)
    if value is None:
        return
    return " ".join(value.split())


@lru_cache(1024)
def clean_name(value: Any) -> str | None:
    """
    Clean a value and only return it if it is a "name" in the sense of, doesn't
    contain exclusively of special chars
    """
    value = clean_string(value)
    if slugify(value) is None:
        return
    return value


@lru_cache(1024)
def fingerprint(value: Any) -> str | None:
    """
    Create a stable but simplified string or None from input
    that can be used to generate ids (to mimic `fingerprints.generate` which is
    unstable for IDs as its algorithm could change)
    """
    value = clean_name(value)
    if value is None:
        return
    return " ".join(sorted(set(slugify(value).split("-"))))


@lru_cache(1024)
def string_id(value: Any) -> str | None:
    return make_entity_id(fingerprint(value))


def str_or_none(value: Any) -> str | None:
    if not value:
        return None
    value = str(value).strip()
    return value or None


def join_text(*parts: Any, sep: str = " ") -> str | None:
    parts = [clean_name(p) for p in parts]
    return _join_text(*parts, sep=sep)


def checksum(io: BytesIO, algorithm: str | None = "md5") -> str:
    handler = getattr(hashlib, algorithm)
    hash_ = handler()
    for chunk in iter(lambda: io.read(128 * hash_.block_size), b""):
        hash_.update(chunk)
    return hash_.hexdigest()


def data_checksum(data: Any, algorithm: str | None = "md5") -> str:
    if is_listish(data):
        data = sorted(data, key=lambda x: repr(x))
    data = orjson.dumps(data, option=orjson.OPT_SORT_KEYS)
    return checksum(BytesIO(data), algorithm)


def is_empty(value: Any) -> bool:
    if isinstance(value, (bool, int)):
        return False
    if value == "":
        return False
    return not value


def dict_merge(d1: dict[Any, Any], d2: dict[Any, Any]) -> dict[Any, Any]:
    # update d1 with d2 but omit empty values
    d1, d2 = clean_dict(d1), clean_dict(d2)
    for key, value in d2.items():
        if not is_empty(value):
            if isinstance(value, dict):
                d1[key] = dict_merge(d1.get(key, {}), value)
            else:
                d1[key] = value
    return d1


def pydantic_merge(m1: BaseModel, m2: BaseModel) -> BaseModel:
    if m1.__class__ != m2.__class__:
        raise ImproperlyConfigured(
            f"Cannot merge: `{m1.__class__.__name__}` with `{m2.__class__.__name__}`"
        )
    return m1.__class__(**dict_merge(m1.dict(), m2.dict()))


def join_slug(
    *parts: str | None,
    prefix: str | None = None,
    sep: str = "-",
    strict: bool = True,
    max_len: int = 255,
) -> str | None:
    sections = [slugify(p, sep=sep) for p in parts]
    if strict and None in sections:
        return None
    texts = [p for p in sections if p is not None]
    if not len(texts):
        return None
    prefix = slugify(prefix, sep=sep)
    if prefix is not None:
        texts = [prefix, *texts]
    return sep.join(texts)[:max_len].strip(sep)


def to_dict(obj: Any) -> dict[Any]:
    if hasattr(obj, "dict"):
        return obj.dict()
    if hasattr(obj, "to_dict"):
        return obj.to_dict()
    return ensure_dict(obj)



"""
Inspect dataset pipelines interactively
"""


from typing import Any, Generator

import pandas as pd
from nomenklatura.entity import CE
from rich import print

from investigraph.logic.extract import extract_pandas
from investigraph.model import Resolver, Source
from investigraph.model.config import Config, get_config
from investigraph.model.context import init_context
from investigraph.util import PathLike


def print_error(msg: str):
    print(f"[bold red]ERROR[/bold red] {msg}")


def get_records(source: Source) -> list[dict[str, Any]]:
    records: list[dict[str, Any]] = []
    print("Fetching `%s` ..." % source.uri)
    res = Resolver(source=source)
    for ix, rec in enumerate(extract_pandas(res)):
        records.append(rec)
        if ix == 5:
            return records


def inspect_config(p: PathLike) -> Config:
    config = get_config(path=p)
    try:
        if not callable(config.extract.get_handler()):
            print_error(f"module not found or not callable: `{config.extract.handler}`")
    except ModuleNotFoundError:
        print_error(f"no custom extract module: `{config.extract.handler}`")
    try:
        if not callable(config.transform.get_handler()):
            print_error(
                f"module not found or not callable: `{config.transform.handler}`"
            )
    except ModuleNotFoundError:
        print_error(f"no custom transform module: `{config.transform.handler}`")
    try:
        if not callable(config.load.get_handler()):
            print_error(f"module not found or not callable: `{config.load.handler}`")
    except ModuleNotFoundError:
        print_error(f"no custom load module: `{config.load.handler}`")
    return config


def inspect_extract(config: Config) -> Generator[tuple[str, pd.DataFrame], None, None]:
    """
    Preview fetched & extracted records in tabular format
    """
    for source in config.extract.sources:
        df = pd.DataFrame(get_records(source))
        yield source.name, df


def inspect_transform(config: Config) -> Generator[tuple[str, CE], None, None]:
    """
    Preview first proxies
    """
    for source in config.extract.sources:
        ctx = init_context(config, source)
        proxies: list[CE] = []
        for ix, rec in enumerate(get_records(source)):
            for proxy in ctx.config.transform.handle(ctx, rec, ix):
                proxies.append(proxy)
        yield source.name, proxies



"""
Seed sources for extraction
"""

from typing import Generator

from fsspec import get_fs_token_paths

from investigraph.model.context import BaseContext
from investigraph.model.source import Source


def handle(ctx: BaseContext) -> Generator[Source, None, None]:
    if ctx.config.seed.glob is not None:
        _, _, uris = get_fs_token_paths(ctx.config.seed.glob)
        for uri in uris:
            yield Source(uri=uri)



import os
from datetime import timedelta
from pathlib import Path
from typing import Any

from banal import as_bool
from ftmstore import settings as ftmstore_settings
from prefect.settings import PREFECT_API_DATABASE_CONNECTION_URL


def get_env(env: str, default: Any | None = None) -> Any | None:
    return os.environ.get(env, default)


VERSION = "0.3.2"

DEBUG = as_bool(get_env("DEBUG", 1))
DATA_ROOT = Path(get_env("DATA_ROOT", Path.cwd() / "data")).absolute()
DATASETS_REPO = "https://github.com/investigativedata/investigraph-datasets.git"
DATASETS_BLOCK = get_env("DATASETS_BLOCK")

DEFAULT_SEEDER = get_env("DEFAULT_SEEDER", "investigraph.logic.seed:handle")
DEFAULT_EXTRACTOR = get_env("DEFAULT_EXTRACTOR", "investigraph.logic.extract:handle")
DEFAULT_TRANSFORMER = get_env(
    "DEFAULT_TRANSFORMER", "investigraph.logic.transform:map_ftm"
)
DEFAULT_LOADER = get_env("DEFAULT_LOADER", "investigraph.logic.load:load_proxies")
DEFAULT_AGGREGATOR = get_env(
    "DEFAULT_AGGREGATOR", "investigraph.logic.aggregate:in_memory"
)

REDIS_URL = get_env("REDIS_URL", "redis://localhost:6379")
REDIS_PREFIX = get_env("REDIS_PREFIX", f"investigraph:{VERSION}")

TASK_CACHE = as_bool(get_env("TASK_CACHE", 1))
TASK_RETRIES = int(get_env("TASK_RETRIES", 3))
TASK_RETRY_DELAY = int(get_env("TASK_RETRY_DELAY", 5))
TASK_CACHE_EXPIRATION = int(get_env("TASK_CACHE_EXPIRATION", 0)) or None  # in minutes
TASK_CACHE_EXPIRATION = (
    timedelta(TASK_CACHE_EXPIRATION) if TASK_CACHE_EXPIRATION is not None else None
)

TASK_RUNNER = get_env("PREFECT_TASK_RUNNER", "").lower()

CHUNK_SIZE = int(get_env("CHUNK_SIZE", 1000))

FTM_STORE_URI = get_env("FTM_STORE_URI", PREFECT_API_DATABASE_CONNECTION_URL.value())
ftmstore_settings.DATABASE_URI = FTM_STORE_URI



"""
hook `requests` into prefect api but let it work like we know:

from investigraph.logic import requests

res = requests.get(url, params={})

"""


from urllib.parse import urlencode

import requests
from prefect import flow, get_run_logger, task
from prefect.tasks import task_input_hash

from investigraph import settings
from investigraph.model.source import Source


def get_request_cache_key(*args, **kwargs) -> str:
    ckey = kwargs.pop("ckey", None)
    if ckey is not None:
        return ckey
    return f"GET#{task_input_hash(*args, **kwargs)}"


@task(
    name="get",
    retries=settings.TASK_RETRIES,
    retry_delay_seconds=settings.TASK_RETRY_DELAY,
    cache_key_fn=get_request_cache_key,
    persist_result=True,
    cache_expiration=settings.TASK_CACHE_EXPIRATION,
    refresh_cache=not settings.TASK_CACHE,
)
def _get(url: str, *args, **kwargs):
    log = get_run_logger()
    kwargs.pop("ckey", None)
    log.info(f"GET {url}?{urlencode(kwargs)}")
    res = requests.get(url, *args, **kwargs)
    assert res.ok
    return res


@flow(name="dispatch-get", flow_run_name="get-{url}", version=settings.VERSION)
def get(url: str, *args, **kwargs):
    """
    Execute `requests.get` within prefect context.
    Do a `head` request beforehand to be able to use cache.
    """
    source = Source(uri=url)
    head = source.head()
    return _get(url, *args, ckey=head.ckey, **kwargs)


# convenience
post = requests.post
head = requests.head
put = requests.put
delete = requests.delete
patch = requests.patch
Response = requests.Response
